{-# LANGUAGE DeriveDataTypeable, CPP #-}
module LibMHA where

import qualified Data.Map as M
import qualified Data.Set as S
import Data.List(insert,group,isPrefixOf)
import Data.Char(isSpace)
import Data.Maybe(fromJust)
import Data.Typeable

import Debug.Trace

#undef TRACE
-- #define TRACE


#ifdef TRACE
trc = trace
#else
trc _ = id
#endif

type Reward = Int
type Moment = (Char,Char,Reward)
newtype FeedbackChopper = FCh {unFCh :: Chopper}
newtype InputChopper    = ICh {unICh :: Chopper} deriving Typeable
type Chopper = Char -> [Moment] -> (Bool, Maybe (String, [Moment]))  -- 外側はillegalな場合にFalse, 内側は読み込み途中にNothing
type RigidChopper = Char -> [Moment] -> Maybe (Maybe (String, [Moment]))
fromChopper :: Chopper -> RigidChopper
fromChopper f inp tups = case f inp tups of (False,_) -> Nothing
                                            (True, m) -> Just m
-- data OutputChopper = ByReward | ByInput deriving (Typeable, Show)
type OutputChopper = Bool
byReward = False
byInput  = True
newtype Retv = R {unR :: String               -- ^ alphabet
                      -> Char                 -- ^ random char
                      -> Char                 -- ^ current input
                      -> [(Char,Char,Reward)] -- ^ [(input,output,reward)] of the current task instance in the reverse order
                      -> (Bool, Maybe Char)}  -- ^ (should not be filtered out, Nothing for skipping to the next candidate and Just for outputting char)
    deriving Typeable
chop :: FeedbackChopper -> InputChopper -> OutputChopper -> Char -> [(Char,Char,Reward)] -> Maybe ((Reflex, [(String, String)]), Maybe (String,String))
chop (FCh feedbackChopper) (ICh inputChopper) outputChopper inp tups
  = trc "chop" $
    case snd (inputChopper inp tups) >>= snd . feedbackChopper inp . snd of   -- 頭出しのためにillegalかどうかをノーチェックでinputChopperとfeedbackChopperを噛ませる
      Nothing          -> Nothing -- return rc         -- Nothing                                         -- 履歴が十分にない場合かchopperが間違っている場合だけど，どうせ履歴が十分にない場合は何を返しても良い（無限ループにさえならなければよい）ので，Nothingにしていた．This should be Nothing, just in case of elem*Chopper [].
      Just (_fb, rest) ->                                                              -- そういう意味ではずっと>>=なりdoなりでダラダラな形で書けるのだが，まあ，どっちが分かりやすいかはよく分からん．
         reflexChopEpisodes (fromChopper feedbackChopper) (fromChopper inputChopper) outputChopper inp rest (M.empty, [])

-- outputmap (a.k.a. conditional reflex)をまずチェックして，すでに経験してあるものであればそのまま返す．
reflexMapChopDo :: FeedbackChopper -- ^ 入力系列を取ってfeedback途中かどうか判断し，途中でなければfeedbackと残りを返す
                -> InputChopper    -- ^ 入力途中かどうか判断アンド前処理．入力途中ならNothing, 入力が完了していれば，Just (前処理した奴)
                -> OutputChopper
                -> ((Char -> Char) -> String -> String -> String) -- ^ wktk (a.k.a. primitive conditioned reflex) x result from the feedbackmap (a.k.a. assoc memory) x input |-> output
                -> Retv           -- ^ current output. Nothing if a chopper detects something wrong.
reflexMapChopDo feedbackChopper inputChopper outputChopper fun
  = R $ \ alphabet rc inp revtups -> -- trace "LibMHA.reflexMapChopDo" $
    case chop feedbackChopper inputChopper outputChopper inp $ reverse revtups of
      Nothing -> (False, Nothing)
      Just ((reflex,assoc), mbTup) ->
          let
--                                                             wktk key = case filter ((==key).input) revtups of (_k,o,r):_ | r>0 -> o
--                                                                                                               _                -> rc
                          wktk k   = wktk' $ filter ((==k).input) revtups ++ [ (undefined, o, 1) | o <- k:alphabet++[' '..'\x7e'] ] -- ++"\x81" ]
                          wktk' []            = rc  -- I guess this should not happen.
                          wktk' ((_k,o,r):ts) | r>0  = o
                                              | r==0 = wktk' ts
                                              | True = wktk' $ filter ((/=o).output) ts
                          -- There used to be the consistency check that returns Nothing when @lookup inputStr assoc@ returns @Just fb@ and @not consistent@. 
                          -- However, 'consistent' should not be used in order to solve A.2.5.6 where reflex must be inconsistent with the program after the environment starts to omit periods in feedbacks. 
                          -- I repeat the same thing in Japanese for my reminder because it is important: consistentでチェックしちゃダメ！ A.2.5.6はプログラムがフェードアウトしてreflexで答えるようになることを要求している！（とまで言うと言い過ぎか？）

{- これでも同じこと、のはず。
                          promised   = and [ fun wktk (fromJust $ lookup key assoc) key == val | (key, Left val) <- M.toList reflex ]
                          consistent = and [ not $ any (\v -> v `eitherIsPrefixOf` fun wktk (fromJust' $ lookup key assoc) key) |  (key, Right vals) <- M.toList reflex ]
                          desired = promised && consistent
-}
                          desired = all (\ (key, eth) -> let result = fun wktk (fromJust' $ lookup key assoc) key in
                                                         trc ("result = "++show result++", eth="++show eth) $
                                                         case eth of Left  val  -> result == val
                                                                     Right vals -> not $ any (\v -> v `eitherIsPrefixOf` result) $ S.toList vals) $ M.toList reflex  -- 途中まで出力して罰を受けることがあり得る一方，短すぎる場合にも対応．
          in trc ("desired="++show desired) $ -- ++" and reflex = "++show (reflex)) $
             if not desired then (False, Nothing) else   -- wktkが不十分な場合って無理？ 今のwktkなら大丈夫なはず．
                 case mbTup of
                      Nothing                     -> (True, Just ' ') -- 入力中かfeedback中
                      Just (inputStr, outputDone) ->
                        case lookup inputStr assoc of
                             Nothing -> case fun wktk [rc] inputStr of
                                          ""        -> (True, Nothing) -- (False, Nothing)でも良さそうだけど一応assocのせいかもしれないし．
                                          theOutput | theOutput `isPrefixOf` outputDone -> (True, Just rc) -- 正解を返したつもりが、「で？」という反応が返ってきて、しょうがなくランダムに行動している感じ．
                                                    | outputDone `isPrefixOf` theOutput -> (True, Just $ head $ drop (length outputDone) theOutput) -- 普通に正解。theOutput==outputDoneの場合は１つ上の条件に当てはまるので、headはfailしない。
                                                    | otherwise -> (True, Just rc)
                                                               -- (False, Just rc)でも良さそうだけど一応assocのせいかもしれないし．
                             Just fb -> trc ("lookup returned Just "++show fb) $
                                        let
                                          theOutput = case M.lookup inputStr reflex of
                                                        Just (Left result) -> result
                                                        _                  -> fun wktk fb inputStr
                                        in case theOutput of "" -> (False, Nothing)
                                                             _ | theOutput `isPrefixOf` outputDone -> (False, Just rc) -- 正解を返したつもりが、「で？」という反応が返ってきて、しょうがなくランダムに行動している感じ
                                                               | outputDone `isPrefixOf` theOutput -> (True, Just $ head $ drop (length outputDone) theOutput) -- 普通に正解。theOutput==outputDoneの場合は１つ上の条件に当てはまるので、headはfailしない。
                                                               | otherwise -> (True, Just rc)



eitherIsPrefixOf              :: (Eq a) => [a] -> [a] -> Bool
eitherIsPrefixOf []     _      =  True
eitherIsPrefixOf _      []     =  True
eitherIsPrefixOf (x:xs) (y:ys) = x == y && eitherIsPrefixOf xs ys

fromJust' (Just x) = x
fromJust' Nothing = error "fromJust': Nothing"

goodEpisode :: [Moment] -> (Bool, Bool)
goodEpisode ms = let rewards = reward $ unzip3 ms
                     noProblem = all (>=0) rewards
                     rewarded  = any (>0) rewards
                 in (noProblem && rewarded, noProblem)  -- 罰がなく，報酬があるエピソード

type Reflex = M.Map String (Either String (S.Set String))

-- 最新の編集中のバージョン
reflexChopEpisodes :: RigidChopper -- ^ feedbackChopper: 入力系列を取ってfeedback途中かどうか判断し，途中でなければfeedbackと残りを返す
                -> RigidChopper -- ^ inputChopper: 入力系列を取ってinput途中かどうか判断し，途中でなければinputと残りを返す
                -> OutputChopper
                -> Char               -- ^ current input
                -> [(Char,Char,Reward)] -- ^ [(input,output,reward)] of the current task instance in the normal order
                -> (Reflex, [(String, String)]) -- ^ (current conditioned reflex, association memory)
                -> Maybe ((Reflex, [(String, String)]), Maybe (String,String)) -- ((next conditional reflex, next association memory), (currentInput, currentOutput))
reflexChopEpisodes feedbackChopper inputChopper outputChopper inp tups maps@(reflex, assoc) = do
      mb <- inputChopper inp tups
      case mb of Nothing            -> return (maps, Nothing) -- 入力中
                 Just (theIn, [])   -> return (maps, Just (theIn,[]))
                 Just (theIn, rrs@(r:rs))
                      | outputChopper == byInput 
                           -> case span (isSpace . input) rs of
                                         (tk,[]) | isSpace inp -> outputting $ r:tk
                                         (tk,dr) -> chopFeedbacks (r:tk) dr 
                      | otherwise           -- byReward
                           -> case span ((==0).reward) rrs of
                                         (tk,[]) -> outputting tk
                                         (tk,d:dr) -> chopFeedbacks (tk++[d]) dr
                   where outputting ts = return (maps, Just (theIn, map output ts))
                         chopFeedbacks tk dr = do 
                                           mb <- feedbackChopper inp dr
                                           case mb of
                                                     Nothing               -> return (maps, Nothing) -- feedback中
                                                     Just (feedback, rest) ->
                                                       let outputMoments = tk
                                                           theOut        = map output outputMoments
                                                           (isGood, noPro) = goodEpisode outputMoments
                                                           newMaps | isGood    = (M.insert theIn (Left theOut) reflex, assoc++[(theIn, feedback)])
                                                                   | otherwise = case M.lookup theIn reflex of Just (Left o) | o == theOut -> (M.singleton theIn (Right $ S.singleton theOut), [(theIn, feedback)])
                                                                                                               Just (Right os) | not noPro -> (M.insert theIn (Right $ S.insert theOut os) reflex, newA)
                                                                                                               Nothing         | not noPro -> (M.insert theIn (Right $ S.singleton theOut) reflex, newA)
                                                                                                               _                           -> (reflex, newA)
                                                             where newA = (theIn, feedback) : assoc
                                                       in reflexChopEpisodes feedbackChopper inputChopper outputChopper inp rest newMaps

isLongerThan :: [a] -> Int -> Bool
isLongerThan xs n = not $ null $ drop n xs

input        (i,_,_) = i
output       (_,o,_) = o
reward       (_,_,r) = r

-- choppers for specific lengths. These should not be used because they are not general.
nullChopper = FCh $ \_i ms -> (True, Just ("", ms))
oneInputChopper :: InputChopper
oneInputChopper = ICh $ \i is -> (True, Just $ case is of []  -> ([i], [])
                                                          j:_ -> ([input j], is))
oneFeedbackChopper :: FeedbackChopper
oneFeedbackChopper = FCh $ \i is -> case is of 
                                              [] -> (True, Nothing)       -- iがfeedbackの終わりの文字であってもまだfeedback中扱い．
                                              m@(j,o,r):js | shouldListen m -> (True, Just ([j], js))
                                                           | otherwise       -> (False, Just (undefined, drop 1 js))
splitInputChopper :: Int -> InputChopper
splitInputChopper n = ICh $ \i is -> let sep  = abs n + 1
                                     in case splitAt sep is of
                                          (tk,dr) | diff >  1 -> (all shouldListen tk, Nothing)
                                                  | diff == 1 -> (True, Just (map input tk ++ [i], []))
                                                  | otherwise -> -- if null tk then error $ "diff = "++show diff ++" and length dr = "++ show (length dr) else
                                                                (all shouldListen $ init tk, Just (map input tk, last tk : dr))
                                                  where diff = sep - length tk

splitFeedbackChopper :: Int -> FeedbackChopper
splitFeedbackChopper n = FCh $ \i is -> case splitAt n is of 
                                          (tk,dr) | length tk < n       -> (all shouldListen tk, Nothing)       -- iがfeedbackの終わりの文字であってもまだfeedback中扱い．
                                                  | all shouldListen tk -> (True,  Just (map input tk, dr))
                                                  | otherwise           -> (False, Just (undefined, drop 1 dr))

spanFeedbackChopper :: (Char -> Bool) -> FeedbackChopper
spanFeedbackChopper cond = FCh $ \i is -> case span (cond . input) is of 
                                               (tk,[])   -> (all shouldListen tk, Nothing)       -- iがfeedbackの終わりの文字であってもまだfeedback中扱い．
                                               (tk,d:dr) -> (all shouldListen tk && shouldListen d, Just (map input tk, dr)) -- dを捨てるべきかtkにつけるべきか？
spanFeedbackChopper' :: (Char -> a -> Bool) -> a -> FeedbackChopper
spanFeedbackChopper' op cs = spanFeedbackChopper (`op` cs)
elemFeedbackChopper :: [Char] -> FeedbackChopper
elemFeedbackChopper cs = spanFeedbackChopper (`notElem` cs)

spanInputChopper :: (Char -> Bool) -> InputChopper
spanInputChopper cond = ICh $ \i is -> case span (cond . input) is of
                                               (tk,[])  | cond i -> (all shouldListen tk, Nothing)
                                               ([],dr) -> (False, Just ("", dr))  -- The input must not be "" in order to avoid infinite loop.
                                               (tk,dr) -> (all shouldListen tk, Just (map input tk, dr))  -- iがinputの終わりの文字ならinput終わり．
spanInputChopper' :: (Char -> a -> Bool) -> a -> InputChopper
spanInputChopper' op cs = spanInputChopper (`op` cs)
elemInputChopper :: [Char] -> InputChopper
elemInputChopper cs = spanInputChopper (`notElem` cs)

shouldListen :: Moment -> Bool
shouldListen (_,o,r) = o==' ' && r==0 || o/=' ' && r<0


minFeedbackChopper :: FeedbackChopper -> FeedbackChopper -> FeedbackChopper
minFeedbackChopper (FCh lc) (FCh rc) = FCh $ minChopper lc rc
minInputChopper    :: InputChopper    -> InputChopper    -> InputChopper
minInputChopper    (ICh lc) (ICh rc) = ICh $ minChopper lc rc

-- minChopper is like mappend where (lin, lms) `mappend` (rin, rms) = if lin<rin then (lin, lms) else (rin, rms)
minChopper :: Chopper -> Chopper -> Chopper
minChopper lc rc i ms = case lc i ms of 
                          (_,  Nothing)            -> rt
                          lt@(lb, Just (lin, lms)) -> case rt of
                                                        (_, Nothing)          -> lt
                                                        (rb, Just (rin, rms)) -> if lin<rin then lt else rt
  where rt = rc i ms

-- total extention of (!!), making the list argument circular
[] !!... _ = ' '
xs !!... n = xs !! (n `mod` length xs)

lookupDefault :: String -> String -> [(String, String)] -> String
lookupDefault d k fm = maybe d id $ lookup k fm

dropUntilElem :: String->String->String
dropUntilElem cs = drop 1 . dropWhile (`notElem` cs)

dropUntilEq :: Char -> String -> String
dropUntilEq c = dropUntilElem [c]

snoc :: String -> Char -> String
snoc xs y = xs++[y]

spanDo :: Char -> (String -> String -> String) -> (String -> String) -> String -> String
spanDo c f g xs = case span (/=c) xs of (tk,_c:dr) -> f tk dr
                                        (_t,[])    -> g xs

spanElemDo :: String -> (String -> String -> String) -> (String -> String) -> String -> String
spanElemDo cs f g xs = case span (`notElem` cs) xs of (tk,c:dr) -> f (tk++[c]) dr
                                                      (_t,[])   -> g xs
