module LibMHAClosure where
import qualified LibMHA as L

type Choppers = (FeedbackChopper -- ^ 入力系列を取ってfeedback途中かどうか判断し，途中でなければfeedbackと残りを返す
                , InputChopper   -- ^ 入力途中かどうか判断アンド前処理．入力途中ならNothing, 入力が完了していれば，Just (前処理した奴)
                , L.OutputChopper)

type FeedbackChopper = ([Either Char Int], Either Char Int)
type InputChopper    = (Either Char Int, [Either Char Int])

splitFeedbackChopper :: Int -> FeedbackChopper
splitFeedbackChopper i = ([], Right i)
splitInputChopper    :: Int -> InputChopper
splitInputChopper    i = (Right i, [])

eqFeedbackChopper  :: Char -> FeedbackChopper
eqFeedbackChopper c = ([], Left c)
eqInputChopper     :: Char -> InputChopper
eqInputChopper    c = (Left c, [])

chop :: Choppers -> Char -> [(Char,Char,L.Reward)] -> Maybe ((L.Reflex, [(String, String)]), Maybe (String,String))
chop (fch, ich, och) = L.chop (fromClosureFCh fch) (fromClosureICh ich) och

reflexMapChopDo :: Choppers
                -> ((Char -> Char) -> String -> String -> String) -- ^ wktk (a.k.a. primitive conditioned reflex) x result from the feedbackmap (a.k.a. assoc memory) x input |-> output
                -> L.Retv           -- ^ current output. Nothing if a chopper detects something wrong.
reflexMapChopDo (fch, ich, och) = L.reflexMapChopDo (fromClosureFCh fch) (fromClosureICh ich) och

fcfch :: Either Char Int -> L.FeedbackChopper
fcfch (Right i) = L.splitFeedbackChopper i
fcfch (Left c)  = L.elemFeedbackChopper [c]
fcich :: Either Char Int -> L.InputChopper
fcich (Right i) = L.splitInputChopper i
fcich (Left c)  = L.elemInputChopper [c]

fromClosureFCh :: FeedbackChopper -> L.FeedbackChopper
fromClosureFCh (es, e)   = foldr L.minFeedbackChopper (fcfch e) $ map fcfch es
fromClosureICh :: InputChopper -> L.InputChopper
fromClosureICh (e, es)   = foldr L.minInputChopper (fcich e) $ map fcich es
-- foldlの方が実感には近いが．

minFeedbackChopper :: FeedbackChopper -> FeedbackChopper -> FeedbackChopper
minFeedbackChopper (ls, l) (rs, r) = (ls++l:rs, r)
minInputChopper :: InputChopper -> InputChopper -> InputChopper
minInputChopper    (l, ls) (r, rs) = (r, ls++l:rs)


-- unnecessary stuff
elemFeedbackChopper :: [Char] -> FeedbackChopper
elemFeedbackChopper = foldr1 minFeedbackChopper . map eqFeedbackChopper
elemInputChopper    :: [Char] -> InputChopper
elemInputChopper    = foldr1 minInputChopper . map eqInputChopper
