{-# LANGUAGE TemplateHaskell, MagicHash, CPP #-}
module MHAgent where

import Data.Char(isAlpha, isPunctuation)
import Data.List
import Data.List.Extra(takeEnd, dropEnd)
import Data.Maybe(fromJust, isJust)
import qualified Data.IntMap as IM
import MagicHaskeller
import MagicHaskeller.LibTH(iF, timeout)
import MagicHaskeller.ProgGenSF(mkTrieOptSFIO)
import MagicHaskeller.CoreLang -- (CoreExpr, primitiveToDynamic, parentsOfLeaves, ceToPriority)
-- import MagicHaskeller.Minimal(simpleEF) -- simpleEF cannot be used when using data Retv. everyACE is used instead.
import MagicHaskeller.ProgramGenerator(Common(..), initCommon, extractCommon)
import qualified MagicHaskeller.PolyDynamic as PD
import MagicHaskeller.TICE(ceToDynamic)
--import MagicHaskeller.TyConLib
--import MagicHaskeller.ReadTHType(thTypeToType)

import System.IO
import Debug.Trace

import LibMHA
import qualified LibMHAFiltrable as F
import qualified LibMHAClosure as C

import Control.Concurrent.MVar
import Control.Monad(unless, when, guard)
import Control.Monad.IO.Class

import Control.Monad.Search.Combinatorial

import Data.Function(on)

import System.Random

import MagicHaskeller.Instantiate(typeToRandomsOrd)



debug = True

data Monitor = Mon {vertical :: Maybe Handle, horizontal :: Maybe Handle}

monitor :: MonadIO m => Monitor -> Char -> Char -> Int -> [(Char,Char,Int)] -> m ()
monitor mon inp out rew revtups = do
        case vertical mon of
          Nothing -> return ()
          Just h  -> liftIO $ hPutStrLn h $ inp : out : show rew
        -- 履歴を横向きに表示。
        case horizontal mon of
          Nothing -> return ()
          Just h  ->
              let (revins, revouts, revrews) = unzip3 revtups
              in liftIO $ do
                hPutStrLn h $ reverse $ inp:revins
                hPutStrLn h $ reverse $ out:revouts
                hPutStrLn h $ concatMap (take 1 . show) $ reverse $ rew:revrews


-- historyが不十分な場合，正解プログラムでも正解を返せず，正解プログラムが見つからない状態になってしまうはず，ってことで，その時点で正解である可能性のあるプログラムの最初のnumberOfTrials本だけしか試さず，これを過ぎたら諦めてランダムな文字を返す．
numberOfTrials = 1000000
boreLimit = 50
maxNumConRew = 10

common :: Common
common = initCommon options{tv1=True,nrands=repeat 20,MagicHaskeller.LibTH.timeout=Just 20000} $ concat instinctEqClosure

instinctDyn = map (map (primitiveToDynamic $ tcl common)) instinctEqClosure

task :: (MonadIO m, RandomGen g) => Monitor -> MVar Bool -> (String -> m ()) -> m String -> Int -> Common -> [[PD.Dynamic]] -> g -> Int -> Int -> IM.IntMap Int -> [(Char,Char,Int)] -> m ()
task mon mv sender receiver time cmn instinctDyn gen tick numEpi hist ts = do
    liftIO $ hPutStrLn stderr $ "Now in task. time = " ++ show time
    inp:_ <- receiver
#ifdef MONITOR
    liftIO $ hPutStrLn stderr $ "In task, input = "++[inp]
#endif
    let
                                          alphabetOftenToRare :: [Char]
                                          alphabetOftenToRare = map (toEnum.fst) $ sortBy (\(_,a) (_,b) -> b `compare` a) $ IM.toList hist
    let (n, nextgen) = randomR (0, 0 `max` length alphabetOftenToRare) gen
--    liftIO $ putStrLn $ "alphabetOftenToRare = "++alphabetOftenToRare ++ ", and n = "++show n
    let out = if inp == ' ' then ('.':alphabetOftenToRare) !! n else ' '
    sender [out]
    r <- receiver
    let rew = read r
        newts = (inp,out,rew):ts
--      putStr $ inp : out : take 1 (show rew) ++ (if odd tick then "\n" else " ")  -- debugger for A.2.5.1
    
    monitor mon inp out rew ts
--    liftIO $ putStrLn $ inp : out : show rew
    
    b <- liftIO $ readMVar mv
    unless b $
--       if tick < 2 || (numEpi <= lenObs && tick < 10) then
       if numEpi <= lenObs || tick < 10 then
           task mon mv sender receiver (time+1) cmn instinctDyn nextgen (tick+1) (if rew == 0 then numEpi else numEpi+1) (increment (fromEnum inp) hist) newts
                                        else let 
                                          vocabulary, primitivess :: [[PD.Dynamic]]
--                                          vocabulary = map (:[]) $ charsToDyns (tcl cmn) alphabetOftenToRare
--                                          vocabulary = map ((:[]) . primitiveToDynamic (tcl cmn) . charToPrimitive) alphabetOftenToRare
                                          vocabulary = [map (primitiveToDynamic (tcl cmn) . charToPrimitive) alphabetOftenToRare]


                                          primitivess = zipAppend vocabulary instinctDyn



                                        in do
                                          liftIO $ hPutStrLn stderr $ "alphabetOftenToRare = " ++ alphabetOftenToRare
                         --                 error $ "Vocabulary = " ++ alphabetOftenToRare

--                                          brain <- mkPGXOpts mkTrieOptSFIO options{tv1=True,nrands=repeat 20,MagicHaskeller.LibTH.timeout=Just 20000} [] [] primitivess [] :: IO ProgGenSF
--                                          brain <- updatePGXOpts mkTrieOptSFIO Nothing [] [] primitivess [] cmn :: IO ProgGenSF
                                          brain <- liftIO (updatePGXOptsFilt 10 mkTrieOptSFIO Nothing [] [] primitivess [] cmn :: IO ProgGenSF)


--                                          let brain = mkPGOpt options{tv1=True,nrands=repeat 20,MagicHaskeller.LibTH.timeout=Just 20000} [] (head primitivess) :: ProgGen


{-
                                          hPutStrLn stderr "koko"
                                          let progss = everyACE brain True :: [[(CoreExpr, Retv)]]
                                              progs  = concat progss
-}


                                          let progs = concat $ unMx $ do
                                                  (chopCE,choppers) <- Mx $ everyACE brain True :: Matrix (CoreExpr, C.Choppers)
                                                  guard $ isJust $ C.chop choppers inp $ reverse ts   -- 本当は次のinpを待ってnewtsを使うべき．
                                                  (bodyCE,body)     <- Mx $ everyACE brain True :: Matrix (CoreExpr, (Char -> Char) -> String -> String -> String)
                                                  return $ (VarName "reflexMapChopDo" :$ chopCE :$ bodyCE, C.reflexMapChopDo choppers body)

--                                          let progss2 = take 30 $ everythingF brain True :: [[(Exp, Maybe Char)]]
--                                          hPrint stderr $ map length progss
--                                          hPrint stderr $ take 300 $ map (pprint . exprToTHExp (vl $ extractCommon brain) . fst) progs
                                          loop mon mv sender receiver nextgen 0 Nothing {- choppersss -} progs progs (time+1) (extractCommon brain) instinctDyn 0 (tick+1) (if rew == 0 then numEpi else numEpi+1) (increment (fromEnum inp) hist) newts

-- Should this be defined in MagicHaskeller.lhs?
charToPrimitive :: Char -> Primitive
charToPrimitive c = (unsafeCoerce# c, LitE $ CharL c, ConT ''Char)
{-
charsToDyns :: TyConLib -> [Char] -> [PD.Dynamic]
charsToDyns tcl cs = let char = thTypeToType tcl $ ConT ''Char
                     in [ PD.unsafeToDyn tcl char c (LitE $ CharL c) | c <- cs]
-}
readReward :: String -> Int
readReward r = case reads r of [(reward,_)] -> reward
                               _            -> 1      -- I just do not want to fail when the environment is buggy.


loop :: (MonadIO m, RandomGen g) => Monitor -> MVar Bool -> (String -> m ()) -> m String -> g -> Int -> Maybe (CoreExpr, Retv) -> [(CoreExpr, Retv)] -> [(CoreExpr, Retv)] -> Int -> Common -> [[PD.Dynamic]] -> Int -> Int -> Int -> IM.IntMap Int -> [(Char,Char,Int)] -> m ()
loop mon mv sender receiver gen bore lastSuccess expFuns curExpFuns time cmn skills numConRew tick numEpi hist tups = do
  liftIO $ hPutStrLn stderr $ "Now in loop. time = " ++ show time
  inp:_ <- receiver
#ifdef MONITOR
  liftIO $ hPutStrLn stderr $ "In loop, input = "++[inp]
#endif
  let alphabet = map toEnum $ IM.keys hist
      newHist = increment (fromEnum inp) hist

      traceTime str = trace ("time = "++show time ++ ", "++str)
--      traceTime _ = id

  if bore > boreLimit && inp==' ' then do
        let out = ' '
        sender [out]
        r <- receiver
        let rew = readReward r
            newBore | rew==0    = bore
                    | otherwise = 0

        monitor mon inp out rew tups

        task mon mv sender receiver (time+1) cmn skills gen 0 0 IM.empty []
   else do
    let
--        (punctuations, notPunctuation)    = partition isPunctuation alphabet
--        puncFirst = (punctuations ++ notPunctuation)
        cEF = maybe curExpFuns (\efun -> efun : deleteBy ((==) `on` fst) efun curExpFuns) lastSuccess
        results = [ (result, expfun) | ((True, result), expfun@(ce,_)) <- zipWith (\n xf@(e, R f) -> (f alphabet (alphabet !!... (tick+n)) inp tups, xf)) [0..] cEF ]
                                --  , isJust result || usesAssoc ce ] -- 不正解でかつassocが使われていないものはfilter outしてしまってOK. と思ってたけど、wktkが不十分で失敗することもあるはず。
        defaultResult = alphabet!!...tick
        candidate 
          = case [ (o, expfun) | (Just o, expfun) <- take numberOfTrials results ] of 
                              []      -> trace "progs exhausted" $
                                         case results of []  -> (defaultResult, head expFuns)                      -- resultsすら空だとどうしようもないんだけど，エラーにしないため．
                                                         (_,r):_ -> (defaultResult, r) -- historyが不十分でどうしようもない場合．
                              t:_ -> t                                                            -- 通常
        filtExpFuns = map snd results
        ((out,currentExpFun@(ce,_fun)), newLastSuccessCandidate, forceSwitch)
          = case lastSuccess of
              Just efun@(ce, R f) -> case f alphabet (alphabet!!...tick) inp tups of
                                       (True, Just out)         -> ((out, efun), lastSuccess, False)
                                       _ | numConRew >= maxNumConRew -> trace "switch??" $
                                                                   ((defaultResult,efun), Nothing, True)
                                               -- うまく行っていた途中で切り替わった場合、無理にcandidateを探そうとせず、
                                               -- ちゃんとswitchする(か、最低限でもFalseでもいいことにする)。
                                               -- そうしないと、ずっとcandidateが見つからないことに。
                                       (True, Nothing)          -> ((defaultResult,efun), Nothing, False)
                                       (False, _)               -> (candidate, Nothing, False)
              _                  -> (candidate, Nothing, False) -- 次のlastSuccessについては，今回罰でなければJust candidateで．
    sender [out]
--    hPutStrLn stderr "bar"
    b <- liftIO $ readMVar mv
    unless b $ do
        r <- receiver
        let rew = readReward r
            newNumConRew | rew > 0 = succ numConRew
                         | rew < 0 = 0 
                         | True    = numConRew
            newLastSuccess | rew >= 0 = Just currentExpFun
                           | otherwise= Nothing
--                           | otherwise= newLastSuccessCandidate -- 現在のlastSuccessが無矛盾である限り、今回たとえ罰が与えられたとしても(assocにヒットしなかっただけかもなので)lastSuccess継続。
-- と思ったけど、やっぱこれがいかん。lastSuccessに選ばれたものは先頭に来るので、ここまで甘くしなくてもよい。
-- ここまで甘くすることによって、不正解のものがなかなか除外されない。
            newBore | (inp==' ' && out==' ' || bore > boreLimit) && rew==0 = bore+1
                    | otherwise                      = 0

        --      putStr $ inp : out : take 1 (show rew) ++ (if odd tick then "\n" else " ")  -- debugger for A.2.5.1
        liftIO $ hPutStrLn stderr $ "ce = " ++ pprint (exprToTHExp (vl cmn) ce)

        monitor mon inp out rew tups

        liftIO $ hPutStrLn stderr $ "rew = "++ show rew ++ " and numConRew = " ++ show numConRew

        if rew < 0 && numConRew >= maxNumConRew || forceSwitch then
--                                       task mon mv sender receiver (time+1) cmn skills 0 0 IM.empty [] -- 学習無しを試す場合。
                                       task mon mv sender receiver (time+1) cmn (updateSkills cmn ce skills) gen 0 0 IM.empty []
            else -- trace ("next loop whose drop 1 newExpFuns is " ++ concatMap (\(e,_) -> pprint (exprToTHExp (vl cmn) e)) (take 1 $ drop 1 newExpFuns))
              if rew < 0 && not (IM.member (fromEnum inp) hist) then -- rew<0 is an indispensable condition. See notes on Aug. 8, 2017.
                task mon mv sender receiver (time+1) cmn skills gen (tick+1) (if rew == 0 then numEpi else numEpi+1) newHist ((inp,out,rew):tups)
              else loop mon mv sender receiver gen newBore newLastSuccess expFuns filtExpFuns (time+1) cmn skills newNumConRew (tick+1) (if rew == 0 then numEpi else numEpi+1)  newHist ((inp,out,rew):tups)

updateSkills cmn ce skills = foldr zipAppend skills $ newSkills cmn ce
newSkills cmn (_:$chop:$body) = [ replicate pri [] ++ [[dyn]]  -- Well, this is stupid.
                                | subexpr <- subexprs chop ++ subexprs body,
                                  not (headIsX subexpr),
                                  let candidateSkillCE = removeAbsents subexpr
                                      dyn = ceToDynamic (tcl cmn) (vl cmn) candidateSkillCE
                                      pri = subtract (arityCE candidateSkillCE) $ pred $ ceToPriority (vpl cmn) candidateSkillCE,
                                  pri >= 0, 
                                  isJust $ typeToRandomsOrd (tcl cmn) (rt cmn) (PD.dynType dyn)
                                ]

-- usesAssoc (_reflexMapChopDo :$ _Chop :$ Lambda (Lambda ce)) = isPresent 0 ce
-- usesAssoc _ = False -- さりげにreflexMapChopDoじゃないやつをfilter out
isPresent n (Lambda ce)    = isPresent (succ n) ce
isPresent n (FunLambda ce) = isPresent (succ n) ce
isPresent n (f :$ x)       = isPresent n f || isPresent n x
isPresent n (X m)          = n == m
isPresent n (FunX m)       = n == m
isPresent _ _              = False

prioritize 0 = 0
prioritize n = pred n

increment :: IM.Key -> IM.IntMap Int -> IM.IntMap Int
increment = IM.alter (Just . maybe 1 succ)

-- Just observe, namely just output ' ', for @lenObs@ episodes at the beginning of the task instance.
lenObs = 2

instinctEqClosure = ($(p [|((,,) :: ([Either Char Int], Either Char Int)
                                    -> (Either Char Int, [Either Char Int])
                                    -> Bool
                                    -> (([Either Char Int], Either Char Int),
                                        (Either Char Int, [Either Char Int]),
                                        Bool),
                            C.reflexMapChopDo :: (([Either Char Int], Either Char Int),
                                                  (Either Char Int, [Either Char Int]),
                                                  Bool)
                -> ((Char -> Char) -> [Char] -> [Char] -> [Char])
                -> Retv,
                   C.splitFeedbackChopper :: Int -> ([Either Char Int], Either Char Int),
                   C.eqFeedbackChopper :: Char -> ([Either Char Int], Either Char Int),
                   C.splitInputChopper :: Int -> (Either Char Int, [Either Char Int]),
                   C.eqInputChopper :: Char -> (Either Char Int, [Either Char Int]),
--                   C.minFeedbackChopper :: ([Either Char Int], Either Char Int) -> ([Either Char Int], Either Char Int) -> ([Either Char Int], Either Char Int),
--                   C.minInputChopper :: (Either Char Int, [Either Char Int]) -> (Either Char Int, [Either Char Int]) -> (Either Char Int, [Either Char Int]),
                   byReward, byInput)|])
         ++ $(p [|([]::[Char], (:"")::Char->[Char], -- notElem :: Char -> [Char] -> Bool,
                   iF::(->) Bool ([Char]->[Char]->[Char]), null::(->) [Char] Bool, all::(Char->Bool) -> (->)[Char] Bool, isAlpha :: (->) Char Bool, isPunctuation :: (->) Char Bool, map :: (Char->Char) -> (->) [Char] [Char], -- filter :: (Char -> Bool) -> (->) [Char] [Char], 
                   take :: Int -> (->) [Char] [Char], drop :: Int -> (->) [Char] [Char],
                   drop 1 :: (->) [Char] [Char],
                   takeEnd :: Int -> (->) [Char] [Char], dropEnd :: Int -> (->) [Char] [Char],
--                   reverse :: [Char] -> [Char], not :: (->) Bool Bool
--                   , dropWhile :: (Char->Bool) -> (->) [Char] [Char]
                   dropUntilEq :: Char -> (->) [Char] [Char]
--                   , dropUntilElem :: [Char] -> (->) [Char] [Char]
                     , snoc :: [Char] -> Char -> [Char]
                  )|])
         ++ $(p [|(
                   0::Int,succ::Int->Int,(,)::[Char]->([Char]->[Char])->([Char],[Char]->[Char]), []::[([Char],[Char]->[Char])], (:)::([Char],[Char]->[Char])->[([Char],[Char]->[Char])]->[([Char],[Char]->[Char])])|]))
         : replicate 94 []
